@@ -32,14 +32,12 @@
 #include <X11/Shell.h>
 #include <X11/Xmd.h>
 
-#ifdef HAVE_MOTIF
+#if TK==TK_MOTIF
 #include <Xm/Xm.h>
 #include <Xm/PushB.h>
 #include <Xm/Form.h>
 #include <Xm/MwmUtil.h>
-#endif
-
-#ifdef HAVE_LIBXAW
+#elif TK==TK_XAW
 #include <X11/Xaw/Command.h>
 #include <X11/Xaw/Paned.h>
 #include <X11/Xaw/Box.h>
@@ -107,14 +105,14 @@ XtResource        resources[]=
          XtRImmediate,  (XtPointer)30           },
    {  "useBell", "UseBell",   XtRBoolean, sizeof(int),
          XtOffsetOf(XrusRec,useBell),
-         XtRImmediate, (XtPointer)True          },
+         XtRImmediate, (XtPointer)False         },
    {  "switchKeys","SwitchKeys",XtRString,sizeof(String),
          XtOffsetOf(XrusRec,switchKeys),
          XtRString,  DefaultSwitchKeys          },
-   {  "toRusKeys","ToRusKeys",XtRString,sizeof(String),
+   {  "toRusKeys","ToRusKeys",XtRString,  sizeof(String),
          XtOffsetOf(XrusRec,toRusKeys),
          XtRString,  DefaultToRusKeys           },
-   {  "toLatKeys","ToLatKeys",XtRString,sizeof(String),
+   {  "toLatKeys","ToLatKeys",XtRString,  sizeof(String),
          XtOffsetOf(XrusRec,toLatKeys),
          XtRString,  DefaultToLatKeys           },
    {  "xmodmap",  "Xmodmap",  XtRString,  sizeof(String),
@@ -144,6 +142,12 @@ XtResource        resources[]=
    {  "capsLockLed","CapsLockLed",XtRInt, sizeof(int),
          XtOffsetOf(XrusRec,capsLockLed),
          XtRImmediate,  (XtPointer)0            },
+   {  "recheckTime","RecheckTime",XtRInt, sizeof(int),
+         XtOffsetOf(XrusRec,recheckTime),
+         XtRImmediate,  (XtPointer)1500         },
+   {  "noFork",   "NoFork",   XtRBoolean, sizeof(Boolean),
+         XtOffsetOf(XrusRec,noFork),
+         XtRImmediate,  (XtPointer)False        },
 };
 
 XrmOptionDescRec  options[]=
@@ -160,6 +164,10 @@ XrmOptionDescRec  options[]=
    {  "+led",       "*led",        XrmoptionNoArg,   "0"    },
    {  "-icon",      "*icon",       XrmoptionNoArg,   "true" },
    {  "+icon",      "*icon",       XrmoptionNoArg,   "false"},
+   {  "-nofork",    "*noFork",     XrmoptionNoArg,   "true" },
+   {  "+nofork",    "*noFork",     XrmoptionNoArg,   "false"},
+   {  "-fork",      "*noFork",     XrmoptionNoArg,   "false"},
+   {  "+fork",      "*noFork",     XrmoptionNoArg,   "true" },
 };
 
 Display        *disp;
@@ -189,7 +197,7 @@ int   Mode=LAT;
 int   NewMode;
 
 Widget   top_level=NULL;
-#if defined HAVE_MOTIF || defined HAVE_LIBXAW
+#if TK!=TK_NONE
 Widget   form_w;
 Widget   switch_button[2];
 #endif
@@ -214,7 +222,15 @@ int   MappingNotifyIgnoreCount=0;
 XtIntervalId   MappingNotifyTimeout=-1;
 XtIntervalId   raise_tmout=-1;
 
-void  ShowSwitchButton(void);
+#if TK!=TK_NONE
+static void  ShowSwitchButton(void);
+#else
+# define ShowSwitchButton()
+#endif
+
+#define MaxLetterKeySym 0x1000
+#define NationalKeySym(ks) (ks>=128 && ks<MaxLetterKeySym)
+#define AsciiKeySym(ks) (ks>0 && ks<128)
 
 /*________________________________________________________________________
 */
@@ -222,13 +238,14 @@ void  ShowSwitchButton(void);
 static int SwappableKey(KeySym *k)
 {
    return(k[0]>0 && k[1]>0 && k[2]>0 && k[3]>0
-       && k[0]<256 && k[1]<256 && k[2]<256 && k[3]<256);
+       && k[0]<MaxLetterKeySym && k[1]<MaxLetterKeySym
+       && k[2]<MaxLetterKeySym && k[3]<MaxLetterKeySym);
 }
 
 static int CapsLockableKey(KeySym *k)
 {
    KeySym k0=k[0],k1=k[1];
-   return((k0>=128 && k1>=128 && k0<256 && k1<256)
+   return((NationalKeySym(k0) && NationalKeySym(k1))
       || (((k0>='a' && k0<='z') || (k0>='A' && k0<='Z'))
           &&
           ((k1>='a' && k1<='z') || (k1>='A' && k1<='Z'))));
@@ -358,7 +375,7 @@ static void InitCapsLockEmu()
    XFreeModifiermap(mod);
 }
 
-#if defined HAVE_MOTIF || defined HAVE_LIBXAW
+#if TK!=TK_NONE
 void  RaiseButton()
 {
    XRaiseWindow(disp,XtWindow(top_level));
@@ -370,12 +387,10 @@ void  RaiseButtonOnTimeout(XtPointer closure,XtIntervalId *id)
    raise_tmout=-1;
    RaiseButton();
 }
-#endif /* HAVE_MOTIF || HAVE_LIBXAW */
 
 /* shows proper indicator button */
 void  ShowSwitchButton(void)
 {
-#if defined HAVE_MOTIF || defined HAVE_LIBXAW
    int   to_manage=-1;
 
    if(!switch_button[0] || !switch_button[1])
@@ -399,9 +414,32 @@ void  ShowSwitchButton(void)
 
       RaiseButton();
    }
-#endif /* HAVE_MOTIF || HAVE_LIBXAW */
 }
 
+void  UnmapHandler(Widget w,XtPointer closure,XEvent *ev,Boolean *cont)
+{
+   *cont=True;
+   if(ev->type==UnmapNotify && AppData.alwaysMapped && AppData.icon)
+   {
+      XtMapWidget(top_level);
+   }
+}
+
+void  VisibilityChange(Widget w,XtPointer closure,XEvent *ev,Boolean *cont)
+{
+   (void)w; (void)closure;
+   *cont=True;
+
+   if(ev->type==VisibilityNotify
+   && ev->xvisibility.state!=VisibilityUnobscured
+   && AppData.alwaysOnTop && raise_tmout==-1)
+   {
+      raise_tmout=XtAppAddTimeOut(app_context,1000,
+                        (XtTimerCallbackProc)RaiseButtonOnTimeout,NULL);
+   }
+}
+#endif /* TK!=TK_NONE */
+
 void  PerformSwitch(void)
 {
    SwitchKeyboard(Mode^MODE);
@@ -481,6 +519,12 @@ void  AddWindow(Window w)
 
    if(AddToWindowCache(w)) /* if the window was not in cache */
    {
+      if(!XtWindowToWidget(disp,w))
+      {
+         /* if it is not our own window, unselect any events */
+         XSelectInput(disp,w,0);
+      }
+
       if(!XGetWindowAttributes(disp,w,&wa))
       {
          DelWindow(w);
@@ -544,6 +588,13 @@ void  AddChildren(Window w)
    }
 }
 
+void  AddWindowOnTimeout(XtPointer closure,XtIntervalId *id)
+{
+   (void)id;
+   DelWindow((Window)closure);
+   AddWindow((Window)closure);
+}
+
 void  SwitchKeysFire(void)
 {
    NewMode^=MODE;
@@ -573,8 +624,8 @@ void  CheckKeymap()
       {
          if(keymap[i] && keymap[i+1] && keymap[i+2] && keymap[i+3])
          {
-            ok_lat|=(keymap[i+2]>127 && keymap[i+2]<256 && keymap[i+3]>127 && keymap[i+3]<256);
-            ok_rus|=(keymap[i+0]>127 && keymap[i+0]<256 && keymap[i+1]>127 && keymap[i+1]<256);
+            ok_lat+=(NationalKeySym(keymap[i+2]) && NationalKeySym(keymap[i+3]));
+            ok_rus+=(NationalKeySym(keymap[i+0]) && NationalKeySym(keymap[i+1]));
          }
       }
    }
@@ -584,7 +635,7 @@ void  CheckKeymap()
       fprintf(stderr,"%s: key mapping is not in proper coding, exiting.\n",program);
       if(first_time)
       {
-         fprintf(stderr,"   You might probably want to load a russian modmap befor xrus starting.\n"
+         fprintf(stderr,"   You might probably want to load a localized modmap before xrus starting.\n"
                         "   Alternatively, you can specify modmap file name on the command line,\n"
                         "   like this:   ./xrus jcuken-koi8-xrus.xmm\n");
       }
@@ -593,12 +644,13 @@ void  CheckKeymap()
          fprintf(stderr,"   This means your modmap has changed to single-mode one.\n"
                         "   If you were doing xmodmap <file> at the time, check your mapping in\n"
                         "   the file. For normal xrus work it is required modmap to have both\n"
-                        "   russian and latin letters\n");
+                        "   national and latin letters\n");
       }
       Mode=LAT;
       exit(1);
    }
-   NewMode=Mode=(Mode&~MODE)|(ok_rus?RUS:LAT);
+
+   NewMode=Mode=(Mode&~MODE)|(ok_rus>ok_lat?RUS:LAT);
    FixNewMode();
    SwitchKeyboard(NewMode);
 
@@ -607,6 +659,7 @@ void  CheckKeymap()
    first_time=0;
 }
 
+static
 void  MappingNotifyTimeoutHandler(XtPointer closure,XtIntervalId *id)
 {
    (void)closure; (void)id;
@@ -614,31 +667,6 @@ void  MappingNotifyTimeoutHandler(XtPointer closure,XtIntervalId *id)
    CheckKeymap();
 }
 
-#if defined HAVE_MOTIF || HAVE_LIBXAW
-void  UnmapHandler(Widget w,XtPointer closure,XEvent *ev,Boolean *cont)
-{
-   *cont=True;
-   if(ev->type==UnmapNotify && AppData.alwaysMapped && AppData.icon)
-   {
-      XtMapWidget(top_level);
-   }
-}
-
-void  VisibilityChange(Widget w,XtPointer closure,XEvent *ev,Boolean *cont)
-{
-   (void)w; (void)closure;
-   *cont=True;
-
-   if(ev->type==VisibilityNotify
-   && ev->xvisibility.state!=VisibilityUnobscured
-   && AppData.alwaysOnTop && raise_tmout==-1)
-   {
-      raise_tmout=XtAppAddTimeOut(app_context,1000,
-                        (XtTimerCallbackProc)RaiseButtonOnTimeout,NULL);
-   }
-}
-#endif /* HAVE_MOTIF || HAVE_LIBXAW */
-
 void  MainLoop(void)
 {
    XEvent   ev;
@@ -648,6 +676,8 @@ void  MainLoop(void)
 
    for(;;)
    {
+      alloca(0);  /* garbage collection */
+
       if(app_context==NULL)
          return;
       XtAppNextEvent(app_context,&ev);
@@ -660,12 +690,9 @@ void  MainLoop(void)
 
       switch(ev.type)
       {
-      case(CreateNotify):
-         AddWindow(ev.xcreatewindow.window);
-         break;
       case(MapNotify):
-         /* just to be sure we know the latest input mask */
-         AddWindow(ev.xmap.window);
+         AddWindow(ev.xcreatewindow.window);
+         XtAppAddTimeOut(app_context,AppData.recheckTime,AddWindowOnTimeout,(XtPointer)ev.xmap.window);
          break;
       case(DestroyNotify):
          DelWindow(ev.xdestroywindow.window);
@@ -753,6 +780,7 @@ void  sig_chld(int sig)
 void  LockScreen()
 {
    pid_t pid;
+   sigset_t nset;
 
    SwitchKeyboard(Mode&~MODE);
 
@@ -760,6 +788,8 @@ void  LockScreen()
    switch(pid=fork())
    {
    case(0):
+      sigemptyset (&nset);
+      sigprocmask (SIG_SETMASK, &nset, NULL);
       execl("/bin/sh","sh","-c",AppData.locker,NULL);
       fprintf(stderr,"%s: cannot execute /bin/sh - %s\n",program,strerror(errno));
       fflush(stderr);
@@ -773,6 +803,58 @@ void  LockScreen()
    }
 }
 
+void  run_xmodmap(const char *file)
+{
+   char *alt_file;
+
+   XSync(disp,False);
+
+try_again:
+   if(access(file,R_OK)==-1)
+   {
+      if(file[0]!='/')
+      {
+         alt_file=alloca(strlen(KEYMAP_DIR)+strlen(file)+2+4);
+         sprintf(alt_file,"%s/%s",KEYMAP_DIR,file);
+         if(access(alt_file,R_OK)==0)
+            file=alt_file;
+         else
+         {
+            strcat(alt_file,".xmm");
+            if(access(alt_file,R_OK)==0)
+               file=alt_file;
+            else
+            {
+               /* backward compatibility */
+               char *x=strstr(file,"-xrus");
+               if(x)
+               {
+                  memmove(x,x+4,strlen(x+4)+1);
+                  goto try_again;
+               }
+            }
+         }
+      }
+   }
+
+   fflush(stderr);
+   switch(fork())
+   {
+   case(-1):
+      fprintf(stderr,"%s: fork() failed - %s\n",program,strerror(errno));
+      break;
+   case(0):
+      /* child */
+      execlp("xmodmap","xmodmap",file,NULL);
+      fprintf(stderr,"%s: execlp(xmodmap) failed - %s\n",program,strerror(errno));
+      fflush(stderr);
+      _exit(1);
+   default:
+      /* parent */
+      ;
+   }
+}
+
 void  AlarmHandler(int sig)
 {
    (void)sig;
@@ -830,14 +912,51 @@ int   main(int argc,char **argv)
    char  err[1024];
    int   scr;
 
+#if 0
    printf("Xrus | Version " VERSION " | Copyright (c) 1995-96 Alexander V. Lukyanov\n");
    fflush(stdout);
+#endif
 
    program=argv[0];
 
+   if(argc==2)
+   {
+      if(!strcmp(argv[1],"--version")
+      || !strcmp(argv[1],"-version")
+      || !strcmp(argv[1],"-v"))
+      {
+         printf("Xrus(keyboard switcher) version " VERSION "\n"
+                "\n"
+                "Copyright (c) 1995-98 Alexander V. Lukyanov (lav@yars.free.net)\n"
+                "This is free software that gives you freedom to use, modify and distribute\n"
+                "it under certain conditions, see COPYING (GNU GPL) for details.\n"
+                "There is ABSOLUTELY NO WARRANTY, use at your own risk.\n");
+         exit(0);
+      }
+      if(!strcmp(argv[1],"--help")
+      || !strcmp(argv[1],"-help")
+      || !strcmp(argv[1],"-h")
+      || !strcmp(argv[1],"-?"))
+      {
+         printf("Usage: %s [OPTIONS] [keymap_file]\n"
+                "\n"
+                " {-|+}autolock          turn on/off screen autolocking\n"
+                " -locker \"locker cmd\"   specify locker command\n"
+                " -timeout sec           set autolock timeout\n"
+                " {-|+}bell              turn on/off bell on keyboard switch\n"
+                " {-|+}icon              turn on/off indicator icon\n"
+                " -led led_no            use the led to indicate alternative mode\n"
+                " +led                   don't use led for indication\n"
+                " -nofork                don't fork to background\n"
+                "\n",program);
+         exit(0);
+      }
+   }
+
    hook_signals();
 
    XtToolkitInitialize();
+   XtSetLanguageProc(NULL,NULL,NULL);
    app_context=XtCreateApplicationContext();
    disp=XtOpenDisplay(app_context,NULL,"xrus",AppClass,options,XtNumber(options),&argc,argv);
    if(disp==NULL)
@@ -847,8 +966,8 @@ int   main(int argc,char **argv)
    }
    if(argc>1 && argv[1][0]=='-')
    {
-      fprintf(stderr,"Usage: %s [{-|+}autolock] [-locker \"locker cmd\"] [-timeout sec]\n"
-                     "       [{-|+}bell] [{-|+}icon] [-led led_no] [+led] [modmap_file]\n",program);
+      fprintf(stderr,"%s: invalid option -- %s\n",program,argv[1]);
+      fprintf(stderr,"%s: Try `%s --help' for more information\n",program,program);
       exit(1);
    }
 
@@ -872,10 +991,10 @@ int   main(int argc,char **argv)
    xrus_check=0;
 
    StartArgs();
-#ifdef HAVE_MOTIF
+#if TK==TK_MOTIF
    AddArg(XmNmwmDecorations,MWM_DECOR_BORDER);
-#elif !defined HAVE_LIBXAW
-   AddArg(XtNwidth,10);
+#elif TK==TK_NONE
+   AddArg(XtNwidth,10); /* no toolkit - set size to avoid 0x0 size */
    AddArg(XtNheight,10);
 #endif
    AddArg(XtNinput,False);
@@ -903,11 +1022,13 @@ int   main(int argc,char **argv)
 
    if(AppData.xmodmap[0])
    {
-      char *cmd=alloca(strlen(AppData.xmodmap)+9);
-      sprintf(cmd,"xmodmap %s",AppData.xmodmap);
-      system(cmd);
+      run_xmodmap(AppData.xmodmap);
    }
 
+   /* let the keymap to settle */
+   MappingNotifyTimeout=XtAppAddTimeOut(app_context,500,
+                                        MappingNotifyTimeoutHandler,NULL);
+
    /* initialize switch keys */
    SwitchKeys.Fire=SwitchKeysFire;
    ParseKeyCombination(AppData.switchKeys,&SwitchKeys,err);
@@ -920,15 +1041,18 @@ int   main(int argc,char **argv)
    PrintParseErrors("toRusKeys",err);
 
 #if !defined(DEBUG) && !defined(DONT_FORK)
-   switch(fork())
+   if(!AppData.noFork)
    {
-   case(0):
-      break;
-   case(-1):
-      perror("fork");
-      exit(1);
-   default:
-      _exit(0);
+      switch(fork())
+      {
+      case(0):
+         break;
+      case(-1):
+         perror("fork");
+         exit(1);
+      default:
+         _exit(0);
+      }
    }
 #endif
 
@@ -942,10 +1066,10 @@ int   main(int argc,char **argv)
 
    XtAppAddActions(app_context,Actions,XtNumber(Actions));
 
-#if defined HAVE_MOTIF || defined HAVE_LIBXAW
+#if TK!=TK_NONE
    XtAddEventHandler(top_level,StructureNotifyMask,False,UnmapHandler,NULL);
 
-#ifdef HAVE_MOTIF
+#if TK==TK_MOTIF
    StartArgs();
    AddArg(XmNresizable,False);
    form_w=XtCreateManagedWidget("form",xmFormWidgetClass,top_level,args,count);
@@ -984,10 +1108,10 @@ int   main(int argc,char **argv)
    XChangeProperty(disp,XtWindow(top_level),ol_decor_del_atom,XA_ATOM,32,
          PropModeReplace,(void*)ol_decor_del,2);
 
-#ifdef HAVE_LIBXAW
+#if TK!=TK_MOTIF && TK!=TK_NONE
 {
    /* we have to emulate motif hints as we don't have Motif */
-   static CARD32 motif_hints[]={0x2,0xffffffff,0x2,0xffffffff,0x39250};
+   static long motif_hints[]={0x2,0xffffffff,0x2,0xffffffff,0x39250};
    Atom motif_hints_atom=XInternAtom(disp,"_MOTIF_WM_HINTS",False);
    XChangeProperty(disp,XtWindow(top_level),motif_hints_atom,motif_hints_atom,32,
          PropModeReplace,(void*)motif_hints,XtNumber(motif_hints));
@@ -1000,9 +1124,7 @@ int   main(int argc,char **argv)
    KeyboardStateInit();
    InitCapsLockEmu();
 
-   CheckKeymap();
-
-#if defined HAVE_MOTIF || defined HAVE_LIBXAW
+#if TK!=TK_NONE
    if(AppData.icon)
       XtMapWidget(top_level);
 
@@ -1032,7 +1154,7 @@ int   main(int argc,char **argv)
       XtSetValues(switch_button[1],args,count);
       XtSetValues(form_w,args,count);
    }
-#endif /* HAVE_MOTIF || HAVE_LIBXAW */
+#endif /* TK!=TK_NONE */
 
    SetAlarm();
    MainLoop();
