@@ -29,6 +29,8 @@ typedef struct
    Boolean  alwaysMapped;
    Boolean  capsLockEmu;
    int      capsLockLed;
+   int      recheckTime;
+   Boolean  noFork;
 } XrusRec;
 
 extern   XrusRec  AppData;
@@ -54,6 +56,7 @@ extern   Widget   switch_button[2];
 
 void  SetAlarm();
 void  LockScreen();
+void  run_xmodmap(const char *file);
 
 void  cleanup();  /* should be called before DestroyApplicationContext */
 
